@@ -71,21 +71,25 @@
 * Includes.
 \******************************************************************************/
 
+/* The configuration header file should be included first. */
+#include "jasper/jas_config.h"
+
 #include <assert.h>
-#if defined(HAVE_FCNTL_H)
+#if defined(JAS_HAVE_FCNTL_H)
 #include <fcntl.h>
 #endif
 #include <stdlib.h>
 #include <stdarg.h>
 #include <stdio.h>
 #include <ctype.h>
-#if defined(HAVE_UNISTD_H)
+#if defined(JAS_HAVE_UNISTD_H)
 #include <unistd.h>
 #endif
-#if defined(WIN32) || defined(HAVE_IO_H)
+#if defined(WIN32) || defined(JAS_HAVE_IO_H)
 #include <io.h>
 #endif
 
+#include "jasper/jas_debug.h"
 #include "jasper/jas_types.h"
 #include "jasper/jas_stream.h"
 #include "jasper/jas_malloc.h"
@@ -168,11 +172,17 @@ static jas_stream_t *jas_stream_create()
 	return stream;
 }
 
+#if 0
+
+/* Obsolete code. */
+
 jas_stream_t *jas_stream_memopen(char *buf, int bufsize)
 {
 	jas_stream_t *stream;
 	jas_stream_memobj_t *obj;
 
+	JAS_DBGLOG(100, ("jas_stream_memopen(%p, %d)\n", buf, bufsize));
+
 	if (!(stream = jas_stream_create())) {
 		return 0;
 	}
@@ -219,6 +229,8 @@ jas_stream_t *jas_stream_memopen(char *buf, int bufsize)
 		jas_stream_close(stream);
 		return 0;
 	}
+	JAS_DBGLOG(100, ("jas_stream_memopen buffer buf=%p myalloc=%d\n",
+	  obj->buf_, obj->myalloc_));
 
 	if (bufsize > 0 && buf) {
 		/* If a buffer was supplied by the caller and its length is positive,
@@ -233,12 +245,144 @@ jas_stream_t *jas_stream_memopen(char *buf, int bufsize)
 	return stream;
 }
 
+#else
+
+/*
+This function will eventually replace jas_stream_memopen.
+If buf is 0 and bufsize > 0:
+	a buffer is dynamically allocated with size bufsize and this buffer is
+	not growable.
+If buf is 0 and bufsize is 0:
+	a buffer is dynamically allocated whose size will automatically grow to
+	accommodate the amount of data written.
+If buf is not 0:
+	bufsize (which, in this case, is not currently allowed to be zero) is
+	the size of the (nongrowable) buffer pointed to by buf.
+*/
+
+jas_stream_t *jas_stream_memopen2(char *buf, size_t bufsize)
+{
+	jas_stream_t *stream;
+	jas_stream_memobj_t *obj;
+
+	JAS_DBGLOG(100, ("jas_stream_memopen2(%p, %zu)\n", buf, bufsize));
+
+	assert((buf && bufsize > 0) || (!buf));
+
+	if (!(stream = jas_stream_create())) {
+		return 0;
+	}
+
+	/* A stream associated with a memory buffer is always opened
+	for both reading and writing in binary mode. */
+	stream->openmode_ = JAS_STREAM_READ | JAS_STREAM_WRITE | JAS_STREAM_BINARY;
+
+	/* Since the stream data is already resident in memory, buffering
+	is not necessary. */
+	/* But... It still may be faster to use buffering anyways. */
+	jas_stream_initbuf(stream, JAS_STREAM_FULLBUF, 0, 0);
+
+	/* Select the operations for a memory stream. */
+	stream->ops_ = &jas_stream_memops;
+
+	/* Allocate memory for the underlying memory stream object. */
+	if (!(obj = jas_malloc(sizeof(jas_stream_memobj_t)))) {
+		jas_stream_destroy(stream);
+		return 0;
+	}
+	stream->obj_ = (void *) obj;
+
+	/* Initialize a few important members of the memory stream object. */
+	obj->myalloc_ = 0;
+	obj->buf_ = 0;
+
+	/* If the buffer size specified is nonpositive, then the buffer
+	is allocated internally and automatically grown as needed. */
+	if (!bufsize) {
+		obj->bufsize_ = 1024;
+		obj->growable_ = 1;
+	} else {
+		obj->bufsize_ = bufsize;
+		obj->growable_ = 0;
+	}
+	if (buf) {
+		obj->buf_ = JAS_CAST(unsigned char *, buf);
+	} else {
+		obj->buf_ = jas_malloc(obj->bufsize_);
+		obj->myalloc_ = 1;
+	}
+	if (!obj->buf_) {
+		jas_stream_close(stream);
+		return 0;
+	}
+	JAS_DBGLOG(100, ("jas_stream_memopen2 buffer buf=%p myalloc=%d\n",
+	  obj->buf_, obj->myalloc_));
+
+	if (bufsize > 0 && buf) {
+		/* If a buffer was supplied by the caller and its length is positive,
+		  make the associated buffer data appear in the stream initially. */
+		obj->len_ = bufsize;
+	} else {
+		/* The stream is initially empty. */
+		obj->len_ = 0;
+	}
+	obj->pos_ = 0;
+	
+	return stream;
+}
+
+/*
+NOTE:
+The version of the function jas_stream_memopen only exists for backwards
+compatibility.
+Eventually, it should be replaced by jas_stream_memopen2.
+In retrospect, it was a very poor choice to have specified the buffer
+size parameter (bufsize) to have type int.  On some machines, int may only
+be a 16-bit integer.  This precludes larger-sized buffer allocations, which
+are needed in practice.
+
+If bufsize <= 0, the buffer is growable; otherwise, the buffer has a fixed
+size of bufsize.
+If buf is 0, the buffer is dynamically allocated with jas_malloc.
+If buf is not 0 and bufsize <= 0 (which is not permitted in any
+circumstances), bad things will happen (especially if the buf was not
+allocated with jas_malloc).
+*/
+jas_stream_t *jas_stream_memopen(char *buf, int bufsize)
+{
+	char *new_buf;
+	size_t new_bufsize;
+
+	JAS_DBGLOG(100, ("jas_stream_memopen(%p, %d)\n", buf, bufsize));
+	if (bufsize < 0) {
+		jas_deprecated("negative buffer size for jas_stream_memopen");
+	}
+	if (buf && bufsize <= 0) {
+		// This was never a valid thing to do with the old API.
+		jas_eprintf("Invalid use of jas_stream_memopen detected.\n");
+		jas_deprecated("A user-provided buffer for "
+		  "jas_stream_memopen cannot be growable.\n");
+	}
+	if (bufsize <= 0) {
+		new_bufsize = 0;
+		new_buf = 0;
+	} else {
+		new_bufsize = bufsize;
+		new_buf = buf;
+	}
+	return jas_stream_memopen2(new_buf, new_bufsize);
+}
+
+#endif
+
 jas_stream_t *jas_stream_fopen(const char *filename, const char *mode)
 {
 	jas_stream_t *stream;
 	jas_stream_fileobj_t *obj;
 	int openflags;
 
+	JAS_DBGLOG(100, ("jas_stream_fopen(\"%s\", \"%s\")\n", filename, mode));
+
 	/* Allocate a stream object. */
 	if (!(stream = jas_stream_create())) {
 		return 0;
@@ -283,6 +427,9 @@ jas_stream_t *jas_stream_fopen(const char *filename, const char *mode)
 
 	/* Open the underlying file. */
 	if ((obj->fd = open(filename, openflags, JAS_STREAM_PERMS)) < 0) {
+		// Free the underlying file object, since it will not otherwise
+		// be freed.
+		jas_free(obj);
 		jas_stream_destroy(stream);
 		return 0;
 	}
@@ -298,6 +445,9 @@ jas_stream_t *jas_stream_freopen(const char *path, const char *mode, FILE *fp)
 	jas_stream_t *stream;
 	int openflags;
 
+	JAS_DBGLOG(100, ("jas_stream_freopen(\"%s\", \"%s\", %p)\n", path, mode,
+	  fp));
+
 	/* Eliminate compiler warning about unused variable. */
 	path = 0;
 
@@ -346,6 +496,8 @@ jas_stream_t *jas_stream_tmpfile()
 	jas_stream_t *stream;
 	jas_stream_fileobj_t *obj;
 
+	JAS_DBGLOG(100, ("jas_stream_tmpfile()\n"));
+
 	if (!(stream = jas_stream_create())) {
 		return 0;
 	}
@@ -398,6 +550,8 @@ jas_stream_t *jas_stream_fdopen(int fd, const char *mode)
 	jas_stream_t *stream;
 	jas_stream_fileobj_t *obj;
 
+	JAS_DBGLOG(100, ("jas_stream_fdopen(%d, \"%s\")\n", fd, mode));
+
 	/* Allocate a stream object. */
 	if (!(stream = jas_stream_create())) {
 		return 0;
@@ -448,9 +602,13 @@ jas_stream_t *jas_stream_fdopen(int fd, const char *mode)
 
 static void jas_stream_destroy(jas_stream_t *stream)
 {
+	JAS_DBGLOG(100, ("jas_stream_destroy(%p)\n", stream));
+
 	/* If the memory for the buffer was allocated with malloc, free
 	this memory. */
 	if ((stream->bufmode_ & JAS_STREAM_FREEBUF) && stream->bufbase_) {
+		JAS_DBGLOG(100, ("jas_stream_destroy freeing buffer %p\n",
+		  stream->bufbase_));
 		jas_free(stream->bufbase_);
 		stream->bufbase_ = 0;
 	}
@@ -459,6 +617,8 @@ static void jas_stream_destroy(jas_stream_t *stream)
 
 int jas_stream_close(jas_stream_t *stream)
 {
+	JAS_DBGLOG(100, ("jas_stream_close(%p)\n", stream));
+
 	/* Flush buffer if necessary. */
 	jas_stream_flush(stream);
 
@@ -510,6 +670,12 @@ int jas_stream_read(jas_stream_t *stream, void *buf, int cnt)
 	int c;
 	char *bufptr;
 
+	JAS_DBGLOG(100, ("jas_stream_read(%p, %p, %d)\n", stream, buf, cnt));
+
+	if (cnt < 0) {
+		jas_deprecated("negative count for jas_stream_read");
+	}
+
 	bufptr = buf;
 
 	n = 0;
@@ -529,6 +695,12 @@ int jas_stream_write(jas_stream_t *stream, const void *buf, int cnt)
 	int n;
 	const char *bufptr;
 
+	JAS_DBGLOG(100, ("jas_stream_write(%p, %p, %d)\n", stream, buf, cnt));
+
+	if (cnt < 0) {
+		jas_deprecated("negative count for jas_stream_write");
+	}
+
 	bufptr = buf;
 
 	n = 0;
@@ -576,6 +748,8 @@ char *jas_stream_gets(jas_stream_t *stream, char *buf, int bufsize)
 	char *bufptr;
 	assert(bufsize > 0);
 
+	JAS_DBGLOG(100, ("jas_stream_gets(%p, %p, %d)\n", stream, buf, bufsize));
+
 	bufptr = buf;
 	while (bufsize > 1) {
 		if ((c = jas_stream_getc(stream)) == EOF) {
@@ -594,6 +768,12 @@ char *jas_stream_gets(jas_stream_t *stream, char *buf, int bufsize)
 int jas_stream_gobble(jas_stream_t *stream, int n)
 {
 	int m;
+
+	JAS_DBGLOG(100, ("jas_stream_gobble(%p, %d)\n", stream, n));
+
+	if (n < 0) {
+		jas_deprecated("negative count for jas_stream_gobble");
+	}
 	m = n;
 	for (m = n; m > 0; --m) {
 		if (jas_stream_getc(stream) == EOF) {
@@ -606,6 +786,12 @@ int jas_stream_gobble(jas_stream_t *stream, int n)
 int jas_stream_pad(jas_stream_t *stream, int n, int c)
 {
 	int m;
+
+	JAS_DBGLOG(100, ("jas_stream_pad(%p, %d, %d)\n", stream, n, c));
+
+	if (n < 0) {
+		jas_deprecated("negative count for jas_stream_pad");
+	}
 	m = n;
 	for (m = n; m > 0; --m) {
 		if (jas_stream_putc(stream, c) == EOF)
@@ -634,6 +820,7 @@ int jas_stream_isseekable(jas_stream_t *stream)
 
 int jas_stream_rewind(jas_stream_t *stream)
 {
+	JAS_DBGLOG(100, ("jas_stream_rewind(%p)\n", stream));
 	return jas_stream_seek(stream, 0, SEEK_SET);
 }
 
@@ -641,6 +828,9 @@ long jas_stream_seek(jas_stream_t *stream, long offset, int origin)
 {
 	long newpos;
 
+	JAS_DBGLOG(100, ("jas_stream_seek(%p, %ld, %d)\n", stream, offset,
+	  origin));
+
 	/* The buffer cannot be in use for both reading and writing. */
 	assert(!((stream->bufmode_ & JAS_STREAM_RDBUF) && (stream->bufmode_ &
 	  JAS_STREAM_WRBUF)));
@@ -674,6 +864,8 @@ long jas_stream_tell(jas_stream_t *stream)
 	int adjust;
 	int offset;
 
+	JAS_DBGLOG(100, ("jas_stream_tell(%p)\n", stream));
+
 	if (stream->bufmode_ & JAS_STREAM_RDBUF) {
 		adjust = -stream->cnt_;
 	} else if (stream->bufmode_ & JAS_STREAM_WRBUF) {
@@ -720,7 +912,7 @@ static void jas_stream_initbuf(jas_stream_t *stream, int bufmode, char *buf,
 			/* The buffer must be large enough to accommodate maximum
 			  putback. */
 			assert(bufsize > JAS_STREAM_MAXPUTBACK);
-			stream->bufbase_ = JAS_CAST(uchar *, buf);
+			stream->bufbase_ = JAS_CAST(jas_uchar *, buf);
 			stream->bufsize_ = bufsize - JAS_STREAM_MAXPUTBACK;
 		}
 	} else {
@@ -978,7 +1170,11 @@ long jas_stream_length(jas_stream_t *stream)
 
 static int mem_read(jas_stream_obj_t *obj, char *buf, int cnt)
 {
-	int n;
+	ssize_t n;
+	assert(cnt >= 0);
+	assert(buf);
+
+	JAS_DBGLOG(100, ("mem_read(%p, %p, %d)\n", obj, buf, cnt));
 	jas_stream_memobj_t *m = (jas_stream_memobj_t *)obj;
 	n = m->len_ - m->pos_;
 	cnt = JAS_MIN(n, cnt);
@@ -987,14 +1183,27 @@ static int mem_read(jas_stream_obj_t *obj, char *buf, int cnt)
 	return cnt;
 }
 
-static int mem_resize(jas_stream_memobj_t *m, int bufsize)
+static int mem_resize(jas_stream_memobj_t *m, size_t bufsize)
 {
 	unsigned char *buf;
 
-	assert(m->buf_);
-	if (!(buf = jas_realloc(m->buf_, bufsize))) {
+	//assert(m->buf_);
+	//assert(bufsize >= 0);
+
+	JAS_DBGLOG(100, ("mem_resize(%p, %zu)\n", m, bufsize));
+	if (!bufsize) {
+		jas_eprintf(
+		  "mem_resize was not really designed to handle a buffer of size 0\n"
+		  "This may not work.\n"
+		);
+	}
+
+	if (!(buf = jas_realloc2(m->buf_, bufsize, sizeof(unsigned char))) &&
+	  bufsize) {
+		JAS_DBGLOG(100, ("mem_resize realloc failed\n"));
 		return -1;
 	}
+	JAS_DBGLOG(100, ("mem_resize realloc succeeded\n"));
 	m->buf_ = buf;
 	m->bufsize_ = bufsize;
 	return 0;
@@ -1002,19 +1211,29 @@ static int mem_resize(jas_stream_memobj_t *m, int bufsize)
 
 static int mem_write(jas_stream_obj_t *obj, char *buf, int cnt)
 {
-	int n;
+	size_t n;
 	int ret;
 	jas_stream_memobj_t *m = (jas_stream_memobj_t *)obj;
-	long newbufsize;
-	long newpos;
+	size_t newbufsize;
+	size_t newpos;
 
+	assert(buf);
+	assert(cnt >= 0);
+
+	JAS_DBGLOG(100, ("mem_write(%p, %p, %d)\n", obj, buf, cnt));
 	newpos = m->pos_ + cnt;
 	if (newpos > m->bufsize_ && m->growable_) {
 		newbufsize = m->bufsize_;
 		while (newbufsize < newpos) {
-			newbufsize <<= 1;
-			assert(newbufsize >= 0);
+			//newbufsize <<= 1;
+			if (!jas_safe_size_mul(newbufsize, 2, &newbufsize)) {
+				JAS_DBGLOG(100, ("new buffer size would cause overflow\n"));
+				return -1;
+			}
 		}
+		JAS_DBGLOG(100, ("mem_write resizing from %d to %zu\n", m->bufsize_,
+		  newbufsize));
+		assert(newbufsize > 0);
 		if (mem_resize(m, newbufsize)) {
 			return -1;
 		}
@@ -1041,15 +1260,16 @@ static int mem_write(jas_stream_obj_t *obj, char *buf, int cnt)
 	if (m->pos_ > m->len_) {
 		m->len_ = m->pos_;
 	}
-assert(ret == cnt);
+	assert(ret == cnt);
 	return ret;
 }
 
 static long mem_seek(jas_stream_obj_t *obj, long offset, int origin)
 {
 	jas_stream_memobj_t *m = (jas_stream_memobj_t *)obj;
-	long newpos;
+	size_t newpos;
 
+	JAS_DBGLOG(100, ("mem_seek(%p, %ld, %d)\n", obj, offset, origin));
 	switch (origin) {
 	case SEEK_SET:
 		newpos = offset;
@@ -1074,8 +1294,11 @@ static long mem_seek(jas_stream_obj_t *obj, long offset, int origin)
 
 static int mem_close(jas_stream_obj_t *obj)
 {
+	JAS_DBGLOG(100, ("mem_close(%p)\n", obj));
 	jas_stream_memobj_t *m = (jas_stream_memobj_t *)obj;
+	JAS_DBGLOG(100, ("mem_close myalloc=%d\n", m->myalloc_));
 	if (m->myalloc_ && m->buf_) {
+		JAS_DBGLOG(100, ("mem_close freeing buffer %p\n", m->buf_));
 		jas_free(m->buf_);
 		m->buf_ = 0;
 	}
@@ -1089,25 +1312,33 @@ static int mem_close(jas_stream_obj_t *obj)
 
 static int file_read(jas_stream_obj_t *obj, char *buf, int cnt)
 {
-	jas_stream_fileobj_t *fileobj = JAS_CAST(jas_stream_fileobj_t *, obj);
+	jas_stream_fileobj_t *fileobj;
+	JAS_DBGLOG(100, ("file_read(%p, %p, %d)\n", obj, buf, cnt));
+	fileobj = JAS_CAST(jas_stream_fileobj_t *, obj);
 	return read(fileobj->fd, buf, cnt);
 }
 
 static int file_write(jas_stream_obj_t *obj, char *buf, int cnt)
 {
-	jas_stream_fileobj_t *fileobj = JAS_CAST(jas_stream_fileobj_t *, obj);
+	jas_stream_fileobj_t *fileobj;
+	JAS_DBGLOG(100, ("file_write(%p, %p, %d)\n", obj, buf, cnt));
+	fileobj = JAS_CAST(jas_stream_fileobj_t *, obj);
 	return write(fileobj->fd, buf, cnt);
 }
 
 static long file_seek(jas_stream_obj_t *obj, long offset, int origin)
 {
-	jas_stream_fileobj_t *fileobj = JAS_CAST(jas_stream_fileobj_t *, obj);
+	jas_stream_fileobj_t *fileobj;
+	JAS_DBGLOG(100, ("file_seek(%p, %ld, %d)\n", obj, offset, origin));
+	fileobj = JAS_CAST(jas_stream_fileobj_t *, obj);
 	return lseek(fileobj->fd, offset, origin);
 }
 
 static int file_close(jas_stream_obj_t *obj)
 {
-	jas_stream_fileobj_t *fileobj = JAS_CAST(jas_stream_fileobj_t *, obj);
+	jas_stream_fileobj_t *fileobj;
+	JAS_DBGLOG(100, ("file_close(%p)\n", obj));
+	fileobj = JAS_CAST(jas_stream_fileobj_t *, obj);
 	int ret;
 	ret = close(fileobj->fd);
 	if (fileobj->flags & JAS_STREAM_FILEOBJ_DELONCLOSE) {
@@ -1124,20 +1355,32 @@ static int file_close(jas_stream_obj_t *obj)
 static int sfile_read(jas_stream_obj_t *obj, char *buf, int cnt)
 {
 	FILE *fp;
+	size_t n;
+	int result;
+	JAS_DBGLOG(100, ("sfile_read(%p, %p, %d)\n", obj, buf, cnt));
 	fp = JAS_CAST(FILE *, obj);
-	return fread(buf, 1, cnt, fp);
+	n = fread(buf, 1, cnt, fp);
+	if (n != cnt) {
+		result = (!ferror(fp) && feof(fp)) ? 0 : -1;
+	}
+	result = JAS_CAST(int, n);
+	return result;
 }
 
 static int sfile_write(jas_stream_obj_t *obj, char *buf, int cnt)
 {
 	FILE *fp;
+	size_t n;
+	JAS_DBGLOG(100, ("sfile_write(%p, %p, %d)\n", obj, buf, cnt));
 	fp = JAS_CAST(FILE *, obj);
-	return fwrite(buf, 1, cnt, fp);
+	n = fwrite(buf, 1, cnt, fp);
+	return (n != JAS_CAST(size_t, cnt)) ? (-1) : cnt;
 }
 
 static long sfile_seek(jas_stream_obj_t *obj, long offset, int origin)
 {
 	FILE *fp;
+	JAS_DBGLOG(100, ("sfile_seek(%p, %ld, %d)\n", obj, offset, origin));
 	fp = JAS_CAST(FILE *, obj);
 	return fseek(fp, offset, origin);
 }
@@ -1145,6 +1388,7 @@ static long sfile_seek(jas_stream_obj_t *obj, long offset, int origin)
 static int sfile_close(jas_stream_obj_t *obj)
 {
 	FILE *fp;
+	JAS_DBGLOG(100, ("sfile_close(%p)\n", obj));
 	fp = JAS_CAST(FILE *, obj);
 	return fclose(fp);
 }
